/*
	GuidanceImpl.cpp

	This file is for implementation of custom guidance algorithms for the REMUS 100 AUV.
	Make changes to this source file and corresponding header file only!

	See GuidanceInterface for overridable getters (used by GuidanceThread).

	Copyright: 		Department of Marine Technology, NTNU
	Author: 		Petter Norgren
	Date created: 	2016.09.09
*/

#include "GuidanceImpl.h"
#include "RemusReconImpl.h"
#include "MathUtil.h"

#include <iostream>
#include "LatLong-UTMconversion.h"
#include <fstream>
#include <sstream>  //for std::istringstream
#include <iterator> //for std::istream_iterator
#include <vector>   //for std::vector
#include <string>
#include <iomanip>	//for std::setprecision


CGuidanceImpl::CGuidanceImpl(CRemusReconImpl *pParent)
	: CGuidanceInterface()
	, m_pParent(pParent)
	, m_bInitialized(false)
	, m_ctdThread(NULL)
	, m_headingCommandRadians(0.0)
	, m_altitudeCommandMeters(0.0)
	, m_speedCommandMeterSec(0.0)
	, m_waypointsLatitudeRadians()
	, m_waypointsLongitudeRadians()
	, m_currentWaypoint(0)
	, m_startWaypoint(0)
	, m_stopWaypoint(0)
	, m_numWaypoints(0)
	, m_direction(1)
	, m_waypointsEnable(1)
	, m_directionControl(1)
	, m_errorIntegral(0)
	, m_lookaheadDistance(1)
	, m_integralGain(0)
	, m_saturationLimit(999999999)
	, m_northOffsetPerAltitude(0)
	, m_eastOffsetPerAltitude(0)
	, m_crossTrackErrorWP(45)
	, m_headingWP(0)
	, m_bMissionComplete(false)
	, m_wpLOSnumber(0)
{
	m_ctdThread = new CtdThread(m_pParent);
	m_ctdThread->startThread();
	//m_sensorFusion = new CSensorFusion(m_pParent);
	initialize();
	m_prevSampleTime = CHPTimer::getInstance()->getAbsTime();
}

CGuidanceImpl::~CGuidanceImpl()
{
	m_ctdThread->stopThread();

	if (m_ctdThread != NULL)
	{
		delete(m_ctdThread);
		m_ctdThread = NULL;
	}
}

/*
	Main guidance algorithm implementation. 
	All used commands must be set upon calling this function.
*/
bool CGuidanceImpl::computeNewCommands(const ReconStatus_t &reconState)
{
	if (m_bInitialized == false)
	{
		std::cout << "CGuidanceImpl::computeNewCommands: Thread not initialized.." << std::endl;
		return m_bInitialized;
	}

	double headingLOS = 0;
	int earthRadius = 6371000;

	CCTDData data;

	// Get newest payload data from CTD sensor
	m_ctdThread->getData(data);

	m_waypointsEnable = useWaypointsOrHeading();

	if  ((m_directionControl == 0 || m_directionControl == 2) && m_pParent->getReconDriver()->isActive())	//Heading control or both
	{
		double headingCable, crossTrackError;

		if (m_pParent->getReconDriver()->isActive() )
			waypointLOS(reconState);

		if (m_pParent->getSensorFusionThread()->isRunning())
		{
			headingCable = m_pParent->getSensorFusionThread()->getCableHeading();
			crossTrackError = m_pParent->getSensorFusionThread()->getCrossTrackError();
		}
		else
		{	
			headingCable = m_headingWP;
			crossTrackError = m_crossTrackErrorWP;
		}

		//LOS guidance
		if (abs(crossTrackError) < 30)
			m_errorIntegral += crossTrackError;
		else
			m_errorIntegral = 0;

		if (m_errorIntegral > m_saturationLimit)
			m_errorIntegral = m_saturationLimit;
		else if (m_errorIntegral < -m_saturationLimit)
			m_errorIntegral = -m_saturationLimit;

		headingLOS = headingCable - atan(crossTrackError/m_lookaheadDistance + m_integralGain * m_errorIntegral);

		//std::cout << "Heading LOS: " << headingLOS*180/PI << "  cross-track: "<< crossTrackError << " P term: " << crossTrackError/m_lookaheadDistance << " I: " <<  m_errorIntegral <<std::endl;
	}
	
	// add offset as a function of altitude to detect cables on sidescan image
	double m_latOffsetRad = m_northOffsetPerAltitude/earthRadius * reconState.Altitude;
	double m_longOffsetRad = m_eastOffsetPerAltitude/earthRadius * reconState.Altitude/cos(reconState.Latitude*PI/180);

	if (m_directionControl == 1 || m_directionControl == 2)	//waypoint control or both
	{
		double latitudeDeviation = abs(reconState.Latitude*PI/180 - m_waypointsLatitudeRadians[m_currentWaypoint - 1] - m_latOffsetRad);
		double longitudeDeviation = abs(reconState.Longitude*PI/180 - m_waypointsLongitudeRadians[m_currentWaypoint - 1] - m_longOffsetRad);
		double deviationToleranceRadians = 0.02/60 * PI/180;					//0.02 minutes

		if (m_numWaypoints > 0)
		{
			if (latitudeDeviation < deviationToleranceRadians && longitudeDeviation < deviationToleranceRadians)
			{
				if (m_currentWaypoint* m_direction < m_stopWaypoint* m_direction)
				{
					m_currentWaypoint += m_direction;
					std::cout << "WP target: " << m_currentWaypoint << " Lat/Long: " << m_waypointsLatitudeRadians[m_currentWaypoint - 1]*180/PI << ", " << m_waypointsLongitudeRadians[m_currentWaypoint-1]*180/PI << std::endl;
				}
				else 
				{
					std::cout << "Final waypoint reached. Disabling override. ReturnToClosest = " << GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "ReturnToClosest", 0) << std::endl;
					m_pParent->getReconDriver()->setReturnToClosesPoint(GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "ReturnToClosest", 0));
					m_bMissionComplete = true;
					m_pParent->getReconDriver()->disable();
					m_numWaypoints = 0;
				}
			}
			else if (m_currentWaypoint > m_stopWaypoint)
			{
					std::cout << "Final waypoint reached. Disabling override. ReturnToClosest = " << GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "ReturnToClosest", 0) << std::endl;
					m_pParent->getReconDriver()->setReturnToClosesPoint(GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "ReturnToClosest", 0));
					m_bMissionComplete = true;
					m_pParent->getReconDriver()->disable();
					m_numWaypoints = 0;
			}
		}
		m_latitudeCommandRadians = m_waypointsLatitudeRadians[m_currentWaypoint - 1] + m_latOffsetRad;
		m_longitudeCommandRadians = m_waypointsLongitudeRadians[m_currentWaypoint - 1] + m_longOffsetRad;
	}

	m_headingCommandRadians = headingLOS*180/PI;	//actuallly in degrees!!

	return true;
}

void CGuidanceImpl::initialize()
{
	char waypointFile[128];
	m_startWaypoint = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "StartWaypoint", 1);
	m_stopWaypoint = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "StopWaypoint", -1);
	m_directionControl = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "DirectionCommandAsWaypoints", 1);
	GetConfigStringFromFile(m_pParent->getConfigFileName(), "Guidance", "WaypointFile", "WaypointFile", waypointFile, sizeof(waypointFile));
	m_lookaheadDistance = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","LookaheadDistance",1);
	m_integralGain = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","IntegralGain",0);
	m_saturationLimit = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","SaturationLimit",999999999);
	m_northOffsetPerAltitude = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","OffsetNorthPerAltitude",0);
	m_eastOffsetPerAltitude = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","OffsetEastPerAltitude",0);
	m_altitudeCommandMeters = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","AltitudeCommand",5);
    m_depthCommandMeters = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","DepthCommand",0);
    m_speedCommandMeterSec = GetConfigDoubleFromFile(m_pParent->getConfigFileName(), "Guidance","SpeedCommand",0);


	if (m_directionControl != 0)
	{
		if ( obtainWaypoints(m_waypointsLatitudeRadians, m_waypointsLongitudeRadians, waypointFile) )
		{
			m_numWaypoints = static_cast<int>(m_waypointsLatitudeRadians.size());
			
			m_currentWaypoint = m_startWaypoint;
			m_stopWaypoint = m_stopWaypoint == -1 ? m_numWaypoints : m_stopWaypoint;

			std::cout << "Waypoints successfully loaded from " << waypointFile <<std::endl;
			std::cout << "Using waypoints: "<< m_startWaypoint << " - " << m_stopWaypoint << std::endl;
			m_direction = m_startWaypoint <= m_stopWaypoint ? 1 : -1;
		
			if ( (m_startWaypoint > m_numWaypoints) || (m_startWaypoint < 1) )
			{
				std::cout << "ERROR: StartWaypoint is outside limits of 1-" << m_numWaypoints << std::endl;
			}
			else if (m_stopWaypoint > m_numWaypoints || m_stopWaypoint < 1)
			{
				std::cout << "ERROR: StopWaypoint is outside the limits of 1-" << m_numWaypoints << std::endl;
			}
			else
			{
				m_bInitialized = true;
			}
		}
		else
		{	
			std::cout << "ERROR: Waypoints could not be loaded from '" << waypointFile << "'"<< std::endl;
		}
	}
}

bool CGuidanceImpl::obtainWaypoints(std::vector<double>& wpLat, std::vector<double>& wpLong, const char* const waypointFile)
{
	bool bReadSuccessfully = false;
	
	if (strstr(waypointFile, ".sos"))
	{
		int curveNumber = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "CurveNumber", 0);
		bReadSuccessfully = readSosiFile(waypointFile, wpLat, wpLong, curveNumber);
		if (!bReadSuccessfully)
			std::cout << "Can't find curve in sosi file" << std::endl;
	}
	else if (strstr(waypointFile, ".dat"))
	{
		bReadSuccessfully = readCoordinates(waypointFile, wpLat, wpLong);
		//std::cout << "Read waypoints successfully" << std::endl;
	}
	else
	{
		std::cout << "ERROR: Waypoint file needs to be .sos or .dat" << std::endl;
	}
	
	return bReadSuccessfully;
}

bool CGuidanceImpl::readSosiFile(const char* FileName, std::vector<double>& wpLat,std::vector<double>& wpLong, int curveNumber)
{	
	bool isCorrectCurve = false;
	bool isCoordinates = false;
	bool isValidCoord = false;
	double UTMEasting, UTMNorthing, origoEasting = 0, origoNorthing = 0;
	double Lat, Long;
	double scale = 1;
	int koordsys;
	int referenceEllipsoid = 23;
	int UTMZoneNum = 22;
	char UTMZone[8];

	int Hemisphere = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "Hemisphere", 1);
	const char * HemisphereChar = Hemisphere == 1 ? "N" : "M";
	
	
	std::stringstream strs;
	strs << curveNumber;
	std::string curveNumberStr = strs.str();

    std::ifstream fin( FileName );
    if( !fin  ) {
		std::cerr << "CGuidanceImpl::readSosiFile: Can't open file " << FileName << std::endl;
		return false;
    }

    std::string line;
    while( std::getline( fin, line )) {
		std::istringstream ss(line);
		std::istream_iterator<std::string> begin(ss), end;

		//putting all the tokens in the vector
		std::vector<std::string> arrayTokens(begin, end); 

		if (isCoordinates)
		{	
			isValidCoord = (arrayTokens.size() == 2 || arrayTokens.size() == 3) && line.find(".")==std::string::npos;
			
			if (isValidCoord)
			{
				UTMNorthing = origoNorthing + atof(arrayTokens[0].c_str())*scale;
				UTMEasting = origoEasting + atof(arrayTokens[1].c_str())*scale;
				//height = atof(arrayTokens[2].c_str())*hscale;		...ENHET-H

				UTMtoLL(referenceEllipsoid, UTMNorthing, UTMEasting, UTMZone, Lat, Long );

				wpLat.push_back(Lat*PI/180);
				wpLong.push_back(Long*PI/180);

				std::cout << "Lat/Long: " << std::setprecision(9) << wpLat.back ()*180/PI << " / " << std::setprecision(9) << wpLong.back()*180/PI << std::endl;
			}
			else
				break;
		}
		else if (isCorrectCurve && line.find("..N")!=std::string::npos)
		{
			isCoordinates = true;
		}
		else if ( strcmp(arrayTokens[0].c_str(), ".KURVE")  == 0 && strcmp(arrayTokens[1].c_str(), (curveNumberStr+":").c_str()) == 0 )
		{
			isCorrectCurve = true;
		}
		else if ( strcmp(arrayTokens[0].c_str(), "...ENHET") == 0 )
		{	
			scale = atof(arrayTokens[1].c_str());
		}
		else if (strcmp(arrayTokens[0].c_str(), "...ORIGO-N") == 0 )
		{
			origoEasting = atof(arrayTokens[1].c_str());
			origoNorthing = atof(arrayTokens[2].c_str());
		}
		else if ( strcmp(arrayTokens[0].c_str(), "...KOORDSYS") == 0 )
		{	
			koordsys = atoi(arrayTokens[1].c_str());
			
			if (koordsys >= 21 && koordsys <= 26)		// EUREF89/WGS84
			{
				UTMZoneNum = koordsys + 10;
				referenceEllipsoid = 23;				//WGS84 ellipsoid
			}
			else if (koordsys >= 31 && koordsys <= 36)	// ED50
			{
				UTMZoneNum = koordsys;
				referenceEllipsoid = 14;				// International ellipsoid
			}
			else
				std::cout << "ERROR: Unknown coordinate system in SOSI file" << std::endl;

			std::string UTMZoneStr = static_cast<std::ostringstream*>( &(std::ostringstream() << UTMZoneNum) )->str();
			UTMZoneStr += HemisphereChar;
			strcpy(UTMZone, UTMZoneStr.c_str());
			UTMZone[sizeof(UTMZone) - 1] = 0;
		}
    }
    fin.close();
	return isCorrectCurve;
}


bool CGuidanceImpl::readCoordinates(const char* FileName, std::vector<double>& North, std::vector<double>& East)
{
	std::ifstream fin( FileName );
    if( !fin  ) {
        std::cerr << "Can't open file " << FileName << std::endl;
		return false;
	}

	double n,e;
	while (fin >> n >> e)
	{
		North.push_back(n);
		East.push_back(e);
		//std::cout << "Lat/Long: " << North.back()*180/PI << ", " << East.back()*180/PI <<std::endl;
	}

	fin.close();
	return true;
}

bool CGuidanceImpl::useWaypointsOrHeading()
{	
	switch (m_directionControl)
	{
	case 0:				//Heading
		return false;
	case 1:				//waypoints
		return true;	
	case 2:				//Choose best
		{
			if (m_pParent->getReconDriver()->isActive())
				return false;
			else
				return true;
		}
	default:
		return true;
	}
}

bool CGuidanceImpl::writeData(char* fileName, double data)
{
	bool isTime;
	CAbsTime TimeStamp = CHPTimer::getInstance()->getAbsTime();

	if (strcmp(fileName,"time")==0)
		isTime = true;
	else
		isTime = false;

	char filePath[128];
	strcpy(filePath, "C:\\REMUS\\data\\Guidance\\05.05.17\\");
	strcat(filePath, fileName);
	strcat(filePath, ".dat");

	std::ofstream file_to_be_appended;
	//file_to_be_appended.precision(8);

	file_to_be_appended.open (filePath, std::ofstream::out | std::ofstream::app);
	
	if (isTime)
		file_to_be_appended << TimeStamp << "\n";
	else
		file_to_be_appended << std::setprecision(10) << data << "\n";

	file_to_be_appended.close();

	return true;
}

void CGuidanceImpl::waypointLOS(const ReconStatus_t &reconState)
{
	//m_waypointsLatitudeRadians, m_waypointsLongitudeRadians
	double smallestDeviation = 1000000;
	double secondSmallestDeviation = 1000000;
	int closestWP = 0;
	int secondClosestWP = 0;
	int earthRadius = 6371000;

	for (unsigned int i = 0; i < m_waypointsLatitudeRadians.size(); i++)
	{
		//std::cout << m_waypointsLatitudeRadians[i]*180/PI << "  " << reconState.Latitude << "  "  << m_waypointsLongitudeRadians[i]*180/PI << "  " << reconState.Longitude<< std::endl;

		double deviation = abs(m_waypointsLatitudeRadians[i] - reconState.Latitude*PI/180) + abs(m_waypointsLongitudeRadians[i] - reconState.Longitude*PI/180);

		if (deviation < smallestDeviation)
		{
			secondSmallestDeviation = smallestDeviation;
			secondClosestWP = closestWP;
			smallestDeviation = deviation;
			closestWP = i;
		}
		else if (deviation < secondSmallestDeviation)
		{
			secondSmallestDeviation = deviation;
			secondClosestWP = i;
		}
	}

	//std::cout << "current wp: " << m_currentWaypoint << std::endl;

	int firstWP, secondWP;

	if (closestWP < secondClosestWP)
	{
		if (m_direction == 1)
		{
			firstWP = closestWP;
			secondWP = secondClosestWP;
		}
		else
		{
			firstWP = secondClosestWP;
			secondWP = closestWP;
		}
	}
	else
	{
		if (m_direction != 1)
		{
			firstWP = closestWP;
			secondWP = secondClosestWP;
		}
		else
		{
			firstWP = secondClosestWP;
			secondWP = closestWP;
		}
	}

	if (closestWP*m_direction > m_currentWaypoint*m_direction && m_pParent->getReconDriver()->isActive())
		m_currentWaypoint = secondWP;

	if (m_directionControl == 0 && !m_currentWaypoint* m_direction < m_stopWaypoint* m_direction)
	{	
		std::cout << "waypointLOS: Final waypoint reached. Disabling override." << std::endl;
		m_pParent->getReconDriver()->setReturnToClosesPoint(GetConfigIntFromFile(m_pParent->getConfigFileName(), "Guidance", "ReturnToClosest", 0));
		m_bMissionComplete = true;
		m_pParent->getReconDriver()->disable();
		m_numWaypoints = 0;
	}

	double X = cos(m_waypointsLatitudeRadians[secondWP])*sin(m_waypointsLongitudeRadians[secondWP]-m_waypointsLongitudeRadians[firstWP]);
	double Y = cos(m_waypointsLatitudeRadians[firstWP])*sin(m_waypointsLatitudeRadians[secondWP])-sin(m_waypointsLatitudeRadians[firstWP])*cos(m_waypointsLatitudeRadians[secondWP])*cos(m_waypointsLongitudeRadians[secondWP]-m_waypointsLongitudeRadians[firstWP]);
	
	double heading = PI/2 - atan2((m_waypointsLatitudeRadians[secondWP] - m_waypointsLatitudeRadians[firstWP]), (m_waypointsLongitudeRadians[secondWP]-m_waypointsLongitudeRadians[firstWP]));
	
	m_crossTrackErrorWP = -((m_waypointsLatitudeRadians[firstWP] - reconState.Latitude*PI/180) * sin(heading) - (m_waypointsLongitudeRadians[firstWP] - reconState.Longitude*PI/180)*cos(heading))*earthRadius;
	m_headingWP = PI/2 - atan2(Y,X);

	if (CHPTimer::getInstance()->getAbsTime() > m_prevSampleTime + 1000000)
	{
		m_wpLOSnumber++;
		m_prevSampleTime = CHPTimer::getInstance()->getAbsTime();
	}

	//std::cout << "heading = " << heading << "   headingWP = " << m_headingWP << std::endl;

	//std::cout <<"wp " << firstWP+1 << "/" << secondWP+1 << " wp heading: " << m_headingWP*180/PI << "   wp cross err: " << m_crossTrackErrorWP << "  RECON head: " << reconState.Heading << std::endl;
}
