/*
	GuidanceThread.cpp

	Copyright: 		Department of Marine Technology, NTNU
	Author: 		Petter Norgren
					Date created: 	2016.09.08
*/

#include "GuidanceThread.h"
#include "RemusReconImpl.h"
#include "GuidanceImpl.h"

#include "CommonCode/Log/LogManager.h"
#include "CommonCode/RTOS abstraction layer/config.h"

CGuidanceThread::CGuidanceThread(CRemusReconImpl *pParent) :
	OAL::CThread("CableTrackerGuidanceThread"),
	m_periodicTimer("GuidanceThreadTimer"),
	m_guidance(NULL),
	m_bInitialized(false), 
	m_bRunning(false),
	m_bInControl(false),
	m_bDirectControlEnable(false),
	m_bAllowReconOnErrors(false),
	m_bDepthControlOnly(false),
	m_bWaypointsEnable(false),
	m_bSpeedAsMsEnable(false),
	m_bAltitudeEnable(false),
	m_pParent(pParent),
	m_startLeg(0),
	m_stopLeg(0),
	m_bOverrideStarted(false),
	m_delayOverride(20000)
{
	m_guidance = new CGuidanceImpl(m_pParent);
}

CGuidanceThread::~CGuidanceThread()
{
	stopThread();

	if(m_guidance != NULL)
	{
		delete(m_guidance);
		m_guidance = NULL;
	}
}

void CGuidanceThread::main()
{
	OAL::CWaitable *periodic;
	ReconStatus_t reconState;

	m_pParent->storeEventString("GuidanceThread: main() started.");

	// Set up periodic event
	m_periodicTimer.setPeriodic(c_period_tread);
	periodic = &m_periodicTimer;

	while (!initialize())
	{
		std::cout << "CableTrackerPlugin::GuidanceThread::initialize() failed. Retrying..." << std::endl;
		wait(1000);
	}

	std::cout << "CableTrackerPlugin::GuidanceThread running..." << std::endl;
	
	wait(m_delayOverride);
	m_pParent->getReconDriver()->enable();

	while( !m_ShutDown.wait(0) )
	{
		//std::cout << "GuidanceThread" << std::endl;
		// Wait for periodic timer to fire
		if(periodic->wait())
		{
			m_pParent->getReconDriver()->getRemoteControlState(reconState);
			
			//std::cout << "current leg: " << reconState.Leg << "  isactive?" << m_pParent->getReconDriver()->isActive() << " isRemoteControlAllowed? "<< reconState.RemoteControlAllowed << std::endl;

			// Enable RECON if we have reached the correct leg
			//if (reconState.Leg >= m_startLeg && reconState.Leg < m_stopLeg && m_pParent->getReconDriver()->isActive())
			if (reconState.Leg >= m_startLeg && reconState.Leg < m_stopLeg && m_pParent->getReconDriver()->isConnected())
			{
				//m_pParent->getReconDriver()->disable();
				
				//Enable full override
				if (!m_bOverrideStarted)
				{
					std::cout << "GuidanceThread: Override started" << std::endl;
					//m_pParent->getReconDriver()->disable();
					m_pParent->getReconDriver()->enable();
					m_bOverrideStarted = true;
				}

				if(!m_pParent->getReconDriver()->isActive() && !m_guidance->isMissionComplete())
                {
					std::cout << "GuidanceThread: Enabling recon driver" << std::endl;
                    m_pParent->getReconDriver()->enable();
                }


				if (m_bDepthControlOnly && !reconState.DepthOnlyControlEnabled)
				{
					m_pParent->getReconDriver()->enableDepthOnlyMode();
				}

				if (m_bDirectControlEnable && !reconState.DirectControlEnabled)
				{
					m_pParent->getReconDriver()->enableDirectControlMode();
				}
			}
			else
            {
                if(m_pParent->getReconDriver()->isActive())
                {
                    //m_pParent->getReconDriver()->disableDepthOnlyMode();
                    //m_pParent->getReconDriver()->disableDirectControlMode();
                }
            }


			// Disable RECON if we have reached the correct leg
			if (reconState.Leg >= m_stopLeg && m_pParent->getReconDriver()->isActive())
			{
				m_pParent->getReconDriver()->disable();
			}

			// Disable RECON on mission end
			if (reconState.MissionState == StateMissionEnd)
			{
				m_pParent->getReconDriver()->disable();
			}

			// Disable RECON on errors
			if (!m_bAllowReconOnErrors)
			{
				if (reconState.MissionState == StateMissionError)
				{
					std::cout << "GuidanceThread: Disabling RECON because of error." << std::endl;
					m_pParent->getReconDriver()->disable();
				}
			}

			// Compute new commands in guidance algorithm
			m_guidance->computeNewCommands(reconState);

			// Send commands on every loop even if command is unchanged to avoid RECON-timeout
			
			if (reconState.Leg >= m_startLeg && reconState.Leg < m_stopLeg && m_pParent->getReconDriver()->isActive())
			//if (m_pParent->getReconDriver()->isActive())
			{
				if (!m_bDirectControlEnable)
				{
					sendZCommand();

					//std::cout << "GuidanceThread: sending z command" << std::endl;

					if (!m_bDepthControlOnly)
					{
						sendDirectionCommand();
						sendSpeedCommand();
					}
				} 
				else
				{
					m_pParent->getReconDriver()->setDirectControl(m_guidance->getThrustCommandPercent(), m_guidance->getRudderCommandPercent(), m_guidance->getSternPlaneCommandPercent());
				}
			}
		}
	}

	m_pParent->getReconDriver()->disableDirectControlMode();
	m_pParent->getReconDriver()->disable();

	std::cout << "CableTrackerPlugin::ComThread stopped..." << std::endl;

	// Log an event to the event log
	m_pParent->storeEventString("ComThread: main() stopped.");
}

bool CGuidanceThread::initialize()
{
	if (m_bInitialized)
	{
		return m_bInitialized;
	}

	m_startLeg = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "StartLeg", 10000);
	m_stopLeg = GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "StopLeg", 0);
	m_bDirectControlEnable = (GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "DirectControlEnable", 0) == 1);
	m_bDepthControlOnly = (GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "DepthControlOnlyEnable", 1) == 1);
	m_bSpeedAsMsEnable = (GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "SpeedCommandAsMetersPerSec", 1) == 1);
	m_bAltitudeEnable = (GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "EnableAltitudeControl", 0) == 1);
	m_bAllowReconOnErrors = (GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "AllowReconOnErrors", 0) == 1);
	m_delayOverride = (GetConfigIntFromFile(m_pParent->getConfigFileName(), "Control", "DelayOverride", 0));

	if (m_stopLeg <= m_startLeg)
	{
		m_pParent->storeEventString("GuidanceThread: StopLeg is larger than or equal to StartLeg.");
		std::cout << "ERROR: CableTrackerPlugin: GuidanceThread::startThread() StopLeg is larger than or equal to StartLeg. Please revise." << std::endl;
	}
	
	//m_bInitialized = m_pParent->getReconDriver()->enable();
    m_bInitialized = true;


	return m_bInitialized;
}

bool CGuidanceThread::startThread()
{
	// Check if module has been initialized, if not, initialize module now
	if(!initialize())
	{
		// Log the event to the event log
		m_pParent->storeEventString("GuidanceThread: Failed to initialize module.");
		std::cout << "ERROR: CableTrackerPlugin: GuidanceThread::startThread() Unable to initialize. Restart PP.exe." << std::endl;
		return false;
	}

	try
	{
		// Start thread
		start();
		m_bRunning = true;

		// Log an event to the event log
		m_pParent->storeEventString("GuidanceThread: Thread started.");
		return m_bRunning;
	}
	catch ( ... )
	{
		m_pParent->storeEventString("GuidanceThread: Exception when trying to start thread.");
		return false;
	}
}

void CGuidanceThread::stopThread()
{
	// Shutdown thread
	(void) shutDown();

	// Log an event to the event log
	m_pParent->storeEventString("GuidanceThread: Thread stopped.");
}

bool CGuidanceThread::sendDirectionCommand()
{
	if (m_guidance->getWaypointsEnable() )
	{
		return m_pParent->getReconDriver()->setCurrentWaypoint(m_guidance->getLatitudeCommandRadians(), m_guidance->getLongitudeCommandRadians());
	} 
	else
	{
		return m_pParent->getReconDriver()->setCommandedHeading(m_guidance->getHeadingCommandRadians());
	}
}

bool CGuidanceThread::sendSpeedCommand()
{
	if (m_bSpeedAsMsEnable)
	{
		return m_pParent->getReconDriver()->setCommandedSpeed(m_guidance->getSpeedCommandMeterSec());
	}
	else
	{
		return m_pParent->getReconDriver()->setCommandedRpm(m_guidance->getSpeedCommandRPM());
	}
}

bool CGuidanceThread::sendZCommand()
{
	if (m_bAltitudeEnable)
	{
		return m_pParent->getReconDriver()->setCommandedAltitude(m_guidance->getAltitudeCommandMeters());
	}
	else
	{
		return m_pParent->getReconDriver()->setCommandedDepth(m_guidance->getDepthCommandMeters());
	}
}