/*
	MathUtil.h

	Copyright: 		Department of Marine Technology, NTNU
	Author: 		Petter Norgren
	Date created: 	2016.09.08
*/

#pragma once

#include <math.h>

#include <Eigen/Dense>

#if !defined(PI)
	#define PI 3.14159265359
#endif

#if !defined(MS2KNOTS)
	#define MS2KNOTS 1.94384449
#endif

#if !defined(KNOTS2MS)
#define KNOTS2MS 0.514444444
#endif

#if !defined(R2D)
	#define R2D (180/PI)
#endif

#if !defined(D2R)
	#define D2R (PI/180)
#endif

typedef Eigen::Matrix<double, 3, 1> vector3_T;
typedef Eigen::Matrix<double, 3, 3> matrix3_T;

inline int sign(double x)
{
	int sign;

	if (x < 0) { sign = -1; }
	else { sign = 1; }

	return sign;
}

inline bool isnan(float x)
{
	volatile float d = x;
	return !(x == d);
}

inline void rotationEuler(matrix3_T &mat, double roll, double pitch, double yaw)
{
	mat = Eigen::AngleAxisd(yaw, vector3_T::UnitZ())
		* Eigen::AngleAxisd(pitch, vector3_T::UnitY())
		* Eigen::AngleAxisd(roll, vector3_T::UnitX());
}

inline void rotationEcefToNed(matrix3_T &mat, double lat, double lon)
{
	mat = Eigen::AngleAxisd(lat, vector3_T::UnitY())
		* Eigen::AngleAxisd(lon, vector3_T::UnitZ());
}

inline double normpdf(double x, double s)
{
	double m = 0;
	static const double inv_sqrt_2pi = 0.3989422804014327;
	double a = (x - m) / s;

	return inv_sqrt_2pi / s * std::exp(- 0.5 * a * a);
}

inline double inf2pipi(double psi)
{
	const double saturation = 1E10;

	// Saturate input
	if ( psi >= saturation)
	{
		psi = saturation;
	} 
	else if (psi <= -saturation)
	{
		psi = -saturation;
	}

	return fmod(psi + sign(psi)*PI, 2*PI) - sign(psi)*PI;
}

inline void wgs2utm(double *x, double *y, double lat, double lon)
{
	double lon0, N, T, C, A, M;
	int nf;

	// Parameters
	static const int a = 6378137;						// semi-major axis of the Earth ellipsoid
	static const double b = 6356752.314245;				// semi-minor axis of the Earth ellipsoid
	static const double e = sqrt(1 - pow(b/a, 2));		// first eccentricity
	static const double eps = pow(e,2)/(1 - pow(e,2));	// squared second eccentricity
	static const double k0 = 0.9996;					// scale on central meridian
	static const int ef = 500000;						// false easting

	lon0 = (floor(lon/6)*6+3)*D2R;						// reference longitude in radians

	if(lat < 0)
	{
		nf = 10000000;									// false northing
	}
	else
	{
		nf = 0;
	}
			
	N = a/sqrt(1-pow(e,2)*pow(sin(lat),2)); // N is the radius of curvature of the earth perpendicular to meridian plane
	T = pow(tan(lat),2);             
	C = eps*pow(cos(lat),2);
	A = (lon-lon0)*cos(lat);                       
	M = a*((1 - pow(e, 2)/4 - 3*pow(e,4)/64 - 5*pow(e,6)/256)* lat	// M: true distance along the central meridian from the equator to lat
		-(3*pow(e,2)/8 + 3*pow(e,4)/32 + 45*pow(e,6)/1024)*sin(2*lat)
		+(15*pow(e,4)/256 + 45*pow(e,6)/1024)*sin(4*lat)
		-(35*pow(e,6)/3072)*sin(6*lat));

	// Computing easting
	*x = ef + k0*N*(A + (1 - T + C)*pow(A, 3)/6 + (5 - 18*T + pow(T,2) + 72*C - 58*eps)*pow(A, 5)/120);

	// Computing northing 
	*y = nf + k0*M + k0*N*tan(lat)*(pow(A, 2)/2 +(5 - T + 9*C + 4*pow(C,2))*pow(A,4)/24 + (61 - 58*T + pow(T,2) + 600*C - 330*eps)*pow(A, 6)/720);
}

inline int modulo(int number, int multiple)
{
	return (int)(multiple*(std::ceil(((float)number)/multiple) - (((float)number)/multiple)));
}
