/*
	ReconDriver.cpp

	References: ReconDriver.cpp from HUGIN SDK.

	Copyright: 		Department of Marine Technology, NTNU
	Author: 		Petter Norgren
	Date created: 	2016.09.09	PN Added critical section lock on shared state.
*/

#include "ReconDriver.h"
#include "RemusReconImpl.h"

#include "CommonCode/RTOS abstraction layer/VersionInfo.h"

#include <iostream>
using std::clog;
using std::endl;
#include <math.h>

#define CMD_TIMEOUT	100	// ms

#if !defined(PI)
#define PI 3.14159265
#endif

CReconDriver::CReconDriver(CRemusReconImpl * pOwner, UINT32 nAddress, int nPort) 
    : m_pOwner(pOwner)
    , OAL::CThread("ReconDriver")
    , m_bActive(false)
    , m_bDirect(false)
    , m_bConnected(false)
	, m_bRunning(false)
	, m_nReconAddress(nAddress)
	, m_nReconPort(nPort)
	, m_CmdAck("ReconDriver")
	, m_mutex("RECON state mutex")
{
	// Initialize state struct
	memset( &m_State, 0, sizeof(m_State) );
}

CReconDriver::~CReconDriver(void)
{
	stopThread();

	if(m_bActive)
	{
		// Disable the Recon interface
		(void) disable();
	}

	// Close socket
	(void) m_Socket.Close();
}

bool CReconDriver::connect()
{
	// Check if already connected
	if(m_bConnected)
		return true;

	// Setup Destination address
	memset(&m_Dest, 0, sizeof(m_Dest));

	m_Dest.sin_family	    = AF_INET;
	m_Dest.sin_addr.s_addr  = m_nReconAddress;
	m_Dest.sin_port		    = htons(static_cast<u_short>(m_nReconPort));

	// Setup local socket
	sockaddr_in local_socket;
	memset(&local_socket, 0, sizeof(local_socket));
	local_socket.sin_family		 = AF_INET;
	local_socket.sin_addr.s_addr = htonl(INADDR_ANY);
	local_socket.sin_port		 = htons(static_cast<u_short>(m_nReconPort));

	// Create UDP socket
	if ( m_Socket.Socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP) == SOCKET_ERROR )
	{
		int err = WSAGetLastError();
		clog << "CReconDriver - Could not create socket: " << m_Socket.GetErrorString(err) << endl;
		return false;
	}

	if( m_Socket.Bind((struct sockaddr *) &local_socket,sizeof(local_socket)) == SOCKET_ERROR )
	{
		int err = WSAGetLastError();
		clog << "CReconDriver - Could not bind to local port: " << m_Socket.GetErrorString(err) << endl;
		return false;
	}

	clog << "CReconDriver - Socket successfully bound to port " << m_nReconPort << endl;
	m_bConnected = true;

	return m_bConnected;
}

bool CReconDriver::startThread()
{
	if (!m_bConnected)
		return false;

	if(m_bRunning)
		return true;

	try
	{
		// Start out thread
		start();	
		m_bRunning = true;
	}
	catch ( ... )
	{
		std::clog << "CReconDriver::enable: Exception when starting thread." << std::endl;
	}

	return m_bRunning;
}

bool CReconDriver::stopThread()
{
	// Stop thread
	shutDown();
	m_bRunning = false;

	return m_bRunning;
}

bool CReconDriver::enable()
{
	if (!m_bConnected)
		return false;

	if(m_bActive)
		return true;

	// Send RECON enable command
	send ("#E,1");	//enables full override
	m_bActive = true;

    return m_bActive;
}

bool CReconDriver::disable()
{
	if (!m_bConnected)
		return false;

	// Send RECON disable command
	send ("#E,0");
	m_bActive = false;

	return true;
}

bool CReconDriver::enableDirectControlMode()
{
	if (!m_bConnected)
		return false;

    if (!m_bActive)
    {
        std::clog << "CReconDriver::enableDirectControlMode: NOT active" << std::endl;
        return false;
    }

	// No further action
	m_bDirect = true;
    return true;
}

bool CReconDriver::disableDirectControlMode()
{
	if (!m_bConnected)
		return false;

    if (!m_bActive)
    {
        std::clog << "CReconDriver::disableDirectControlMode: NOT active" << std::endl;
        return false;
    }

	// No further action
	m_bDirect = false;
	return true;
}

bool CReconDriver::enableDepthOnlyMode()
{
	if (!m_bActive)
		return false;

	// Send RECON depth only enable command
	send ("#D,1");
	return true;
}

bool CReconDriver::disableDepthOnlyMode()
{
	if (!m_bActive)
		return false;

	// Send RECON depth only disable command
	send ("#D,0");
	return true;
}

bool CReconDriver::setCommandedHeading ( double Heading )
{
	if (!m_bActive)
		return false;

	if (m_bDirect)
	{
		std::clog << "CReconDriver::setCommandedHeading: NOT allowed in DIRECT mode" << std::endl;
		return false;
	}

	char b[64];
	sprintf( b, "#C,Heading,Goal,%f", Heading );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setCommandedAltitude ( double Altitude )
{
	if (!m_bActive)
		return false;

	if (m_bDirect)
	{
		std::clog << "CReconDriver::setCommandedDepth: NOT allowed in DIRECT mode" << std::endl;
		return false;
	}

	char b[64];
	sprintf( b, "#C,Depth,Altitude,%f", Altitude );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setCommandedDepth ( double Depth )
{
	if (!m_bActive)
		return false;

	if (m_bDirect)
	{
		std::clog << "CReconDriver::setCommandedDepth: NOT allowed in DIRECT mode" << std::endl;
		return false;
	}

	char b[64];
	sprintf( b, "#C,Depth,Depth,%f", Depth );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setCommandedSpeed ( double Speed )
{
	if (!m_bActive)
		return false;

    if (m_bDirect)
    {
        std::clog << "CReconDriver::setCommandedSpeed: NOT allowed in DIRECT mode" << std::endl;
        return false;
    }

	char b[64];
	sprintf( b, "#C,Speed,%f meters/sec", Speed );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setCommandedRpm( double Rpm )
{
	if (!m_bActive)
		return false;

	char b[64];
	sprintf( b, "#C,Speed,%d rpm", (int) Rpm );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setDirectControl ( double Thrust, double Rudder, double Sternplane )
{
	if (!m_bActive)
		return false;

    if (!m_bDirect)
    {
        std::clog << "CReconDriver::setDirectControl: ONLY allowed in DIRECT mode" << std::endl;
        return false;
    }

	char b[64];
	sprintf( b, "#C,Direct,Quadfin,%d,%d,%d", (int)Thrust, (int)Rudder, (int)Sternplane );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setDirectControl ( double Thrust, double Rudder, double Port, double Starboard )
{
	if (!m_bActive)
		return false;

    if (!m_bDirect)
    {
        std::clog << "CReconDriver::setDirectControl: ONLY allowed in DIRECT mode" << std::endl;
        return false;
    }

	char b[64];
	sprintf( b, "#C,Direct,Trifin,%d,%d,%d,%d", (int)Thrust, (int)Rudder, (int)Port, (int)Starboard );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setDirectControlFwd (double Rudder, double Port, double Starboard)
{
	if (!m_bActive)
		return false;

	if (!m_bDirect)
	{
		std::clog << "CReconDriver::setDirectControl: ONLY allowed in DIRECT mode" << std::endl;
		return false;
	}

	char b[64];
	sprintf( b, "#C,Direct,Forward,%d,%d,%d", (int)Rudder, (int)Port, (int)Starboard  );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setCurrentWaypoint ( double Latitude, double Longitude )
{
	if (!m_bActive)
		return false;

    if (m_bDirect)
    {
        std::clog << "CReconDriver::setCurrentWaypoint: NOT allowed in DIRECT mode" << std::endl;
        return false;
    }

	// Get Hemisphere
	char cLat = Latitude  >= 0.0 ? 'N' : 'S';
	char cLon = Longitude >= 0.0 ? 'E' : 'W';

	// Convert to degrees and make absolute
	Latitude  = fabs(Latitude*180.0/PI);
	Longitude = fabs(Longitude*180.0/PI);

	int nLatDeg = (int)Latitude;
	int nLonDeg = (int)Longitude;
	double dLatMin = (Latitude-nLatDeg)*60;
	double dLonMin = (Longitude-nLonDeg)*60;

	char b[64];
	sprintf( b, "#C,Heading,Destination,%d%c%f %d%c%f", nLatDeg, cLat, dLatMin, nLonDeg, cLon, dLonMin  );
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}

bool CReconDriver::setReturnToClosesPoint( int ReturnToClosest)
{
	if (!m_bActive)
		return false;

    if (m_bDirect)
    {
        std::clog << "CReconDriver::setCurrentWaypoint: NOT allowed in DIRECT mode" << std::endl;
        return false;
    }

	char b[64];
	sprintf( b, "#c,%d", ReturnToClosest);
	send( b );
	return m_CmdAck.wait(CMD_TIMEOUT);
}


bool CReconDriver::getRemoteControlState (ReconStatus_t & stat)
{
	if (!m_bActive)
		return false;

	OAL::CCriticalSectionLock Lock(m_mutex);

	stat = m_State;
	return true;
}

//only for simulation
bool CReconDriver::setRemoteControlState (ReconStatus_t & stat)
{
	if (!m_bActive)
		return false;

	OAL::CCriticalSectionLock Lock(m_mutex);

	m_State = stat;
	return true;
}

void CReconDriver::main ( void )
{
	char pBuffer[512];
	fd_set readfds, writefds, exceptfds;
	struct timeval timeout = {1,0};	// 1 sec
	int status;
	int err;
	struct sockaddr_in sFrom;
	int nAddrLen = sizeof(sFrom);

	// Zero non-used socket sets
	FD_ZERO(&writefds);
	FD_ZERO(&exceptfds);

	while( !m_ShutDown.wait(0) && m_bConnected )
	{
		// Request version unless obtained
		if (!m_bGotVersion)
			send( "#v");

		// Initialize the socket sets 
		FD_ZERO(&readfds);

		// Add this socket to the readfds set (must be done before each call to select())
		#pragma warning(push)
		#pragma warning(disable:4127)
		FD_SET(m_Socket.m_sock, &readfds);
		#pragma warning(pop)

		// Wait for event
		status = select(0,&readfds,&writefds,&exceptfds,&timeout);

		if (status > 0)
		{ 
			// Select released. We can now call RecvFrom to get the data. The call is guaranteed not to block.  
			int nLen = m_Socket.RecvFrom(pBuffer,sizeof(pBuffer),0,(struct sockaddr *) &sFrom,&nAddrLen);

			// Check return status
			if(nLen == SOCKET_ERROR)
			{
				err = WSAGetLastError();
				clog << "CReconDriver: Failed to read from the socket. Error code: " << err << endl;

				if( (err == WSAECONNABORTED) || (err == WSAECONNRESET) )
				{
					clog << "CReconDriver: Unrecoverable socket error. Closing socket." << endl;
					m_Socket.Close();
					m_bConnected = false;
					m_bActive	 = false;
					m_pOwner->stopped();
				}
			}
			else if (!verifyMsg(pBuffer, nLen))
			{
				clog << "CReconDriver: Invalid RECON msg, len: " << nLen << ", msg: " << pBuffer << endl;
			}
			else
			{
				handleMsg( pBuffer );
			}
		}
		else if(status == SOCKET_ERROR)
		{
			err = WSAGetLastError();
			clog << "CReconDriver: select() returned error: " << err << endl;
		}
	}
}

void CReconDriver::send(const char * szStr )
{
	char pBuffer[256];
	char pChksum[16];

	if (!m_bConnected)
		return;

	sprintf_s(pBuffer,sizeof(pBuffer),"%s,*",szStr); //concatenate
	// Add Checksum
	int chksum=0;         // declare and init checksum to 0
	for(size_t i=0;i<strlen(pBuffer);i++)       
		chksum += pBuffer[i];// add to checksum
		
	sprintf_s(pChksum,sizeof(pChksum),"%02X\n", 0x0FF & chksum); //concatenate
	strcat_s (pBuffer,sizeof(pBuffer),pChksum);

	// Send
	if (m_Socket.SendTo((char *) pBuffer, strlen(pBuffer), 0, (struct sockaddr *) &m_Dest, sizeof(m_Dest)  ) == SOCKET_ERROR)
	{
		int err = WSAGetLastError();
		clog << "CReconDriver: send failed with error: "  << err << endl;
	}
}

bool CReconDriver::verifyMsg(char * pBuffer, int nLen)
{
	if ((nLen < 6 ) || (*pBuffer != '#'))
		return false;

	char *cptr;
	int chksum;
	char sum;

	// Insert terminating zero
	pBuffer[nLen] = '\0';

	// Need to find the “,*FF\n” sequence at the end of the message.
	cptr = strrchr(pBuffer, ','); // find the last comma
	if (NULL == cptr)
		return false;  //invalid message, no ‘,’ before chksum
	if (cptr[1] != '*')
		return false; //invalid message, no ‘*’

	// Two '*'s accepted as checksum
	if ((cptr[2] == '*') && (cptr[3] == '*'))
	{
		// Remove tail
		*cptr = '\0';
		return true;
	}

	sum = 0;
	do
	{
		sum = sum + *pBuffer;
	}
	while(pBuffer++ != (cptr+1));

	if (1 != sscanf(pBuffer, "%02X", &chksum))
		return false;

	// Remove tail
	*cptr = '\0';

	return (chksum == (sum & 0x0FF));
}

void CReconDriver::sendVersion()
{
	OAL::CVersionInfo vinfo;
	char buf[64];
	sprintf( buf, "#V,%s,%s", vinfo.getProductVersion(), vinfo.getBuildTime() );
	send( buf );
}

double CReconDriver::decodeLatitude(char * pBuffer )
{
	char * p;
	int deg = strtol(pBuffer, &p, 10);
	if (*p++ == 'S')
		return -deg - atof(p)/60;
	else
		return deg + atof(p)/60;
}

double CReconDriver::decodeLongitude(char * pBuffer )
{
	char * p;
	int deg = strtol(pBuffer, &p, 10);
	if (*p++ == 'W')
		return -deg - atof(p)/60;
	else
		return deg + atof(p)/60;
}

void CReconDriver::decodeStateMsg(char * pBuffer)
{
	char * p = strtok( pBuffer, "," );

	OAL::CCriticalSectionLock Lock(m_mutex);

	while (NULL != p)
	{
		if (!strncmp(p,"LAT",3))
			m_State.Latitude = decodeLatitude(p+3);
		else if (!strncmp(p,"LON",3))
			m_State.Longitude = decodeLongitude(p+3);
		else if (!strncmp(p,"DG",2))
			m_State.DepthRef = atof(p+2);
		else if (*p == 'D')
			m_State.Depth = atof(p+1);
		else if (!strncmp(p,"AG",2))
			m_State.AltitudeRef = atof(p+2);
		else if (*p == 'A')
			m_State.Altitude = atof(p+1);
		else if (*p == 'P')
			m_State.Pitch = atof(p+1);
		else if (*p == 'R')
			m_State.Roll = atof(p+1);
		else if (!strncmp(p,"TRG",3))
			m_State.RpmRef = atof(p+3);
		else if (!strncmp(p,"TR",2))
			m_State.Rpm = atof(p+2);
		else if (!strncmp(p,"HG",2))
			m_State.HeadingRef = atof(p+2);
		else if (!strncmp(p,"HR",2))
			m_State.HeadingRate = atof(p+2);
		else if (*p == 'H')
			m_State.Heading = atof(p+1);
		else if (*p == 'V')
			m_State.Speed = atof(p+1);
		else if (*p == 'M')
		{
			int b = strtol(p+1, NULL, 16);
			m_State.MissionState                = (RemusMissionStates)(b & 0x03);
			m_State.RemoteControlEnabled        = (b & 0x04) != 0;
			m_State.RemoteControlActive         = (b & 0x08) != 0;
			m_State.OnShorePower                = (b & 0x10) != 0;
			m_State.ReturnToClosestPointOnLine  = (b & 0x20) != 0;
			m_State.DepthOnlyControlEnabled     = (b & 0x40) != 0;
			m_State.RemoteControlAllowed        = (b & 0x80) == 0;
		}
		else if (*p == 'l')
			m_State.Leg = strtol(p+1, NULL, 10);
		else
		{
			clog << "CReconDriver: Unknown State parameter: "<< p << endl;  // Skip comma
			//std::cout << "CReconDriver: Unknown State parameter: "<< p << endl;
		}

		p = strtok(NULL, ",");
	}
}

void CReconDriver::decodeVersion( char * pBuffer )
{
	char *p;
	m_nVerMaj = strtol(pBuffer,&p, 10);
	m_nVerMin = strtol(p+1,NULL, 10);
	m_bGotVersion = true;

	clog << "CReconDriver: Version: "<< ++pBuffer << endl;  // Skip comma
}

bool CReconDriver::decodeCmdAck( char * pBuffer )
{
	char * p = strtok( pBuffer, "," );
	if ((NULL == p) || (strcmp(p, "Direct") != 0))
		// We only need the Direct values
		return false;
	p = strtok(NULL, ","); if (NULL == p) return false;
	if (strcmp(p,"Trifin" ) == 0)
	{
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Thrust    - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Rudder    - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Port      - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Starboard - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Thrust
		m_nThrust = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Rudder
		m_nPlane[0] = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Port
		m_nPlane[1] = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Starboard
		m_nPlane[2] = atoi(p);
	}
	else if (strcmp(p,"Quadfin" ) == 0)
	{
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Thrust     - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Rudder     - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Sternplane - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Thrust
		m_nThrust = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Rudder
		m_nPlane[0] = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Sternplane
		m_nPlane[1] = atoi(p);
	}
	else if (strcmp(p,"Forward" ) == 0)
	{
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Rudder    - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Port      - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Starboard - ignore
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Rudder
		m_nPlane[3] = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Port
		m_nPlane[4] = atoi(p);
		p = strtok(NULL, ","); if (NULL == p) return false; // p = Actual Starboard
		m_nPlane[5] = atoi(p);
	}
	else
		return false;

	return true;
}

void CReconDriver::handleMsg(char * pBuffer)
{
	char nCmd = *++pBuffer;   // Skip past #
	pBuffer +=2;              // Skip cmd id and comma

	switch (nCmd)
	{
	case 'C':
		decodeCmdAck( pBuffer);
		m_CmdAck.pulse();
		clog << "CReconDriver: Cmd ack: " << pBuffer << endl;
		break;
	case 'E':
		clog << "CReconDriver: Error: " << pBuffer << endl;
		break;
	case 'S':
		decodeStateMsg( pBuffer);
		break;
	case 'v':
		sendVersion();
		break;
	case 'V':
		decodeVersion( pBuffer );
		break;
	default:
		clog << "CReconDriver: Msg ignored: " << nCmd << endl;
	}
}









