/*
	ReconDriver.h

	References: ReconDriver.h from HUGIN SDK.

	Copyright: 		Department of Marine Technology, NTNU
	Author: 		Petter Norgren
	Date created: 	2016.09.09	PN Added critical section lock on shared state.
*/

#pragma once
#include "CommonCode/RTOS abstraction layer/CriticalSection.h"
#include "CommonCode/RTOS abstraction layer/Thread.h"
#include "CommonCode/RTOS abstraction layer/Event.h"
#include "CommonCode/Drivers/HPTimer.h"
#include "CommonCode/Socket/CBerkeleySocket.h"

class CRemusReconImpl;

enum RemusMissionStates { StatePreLaunch, StateInMission, StateMissionError, StateMissionEnd };

typedef struct  
{
	bool RemoteControlAllowed;
	bool RemoteControlEnabled;
	bool RemoteControlActive;
	bool DepthOnlyControlEnabled;
	bool DirectControlEnabled;
	bool ReturnToClosestPointOnLine;
	bool OnShorePower;
	double  Latitude;
	double  Longitude;
	double  Depth;
	double  DepthRef;
	double  Altitude;
	double  AltitudeRef;
	double  Pitch;
	double  Roll;
	double  Rpm;
	double  RpmRef;
	double  Speed;
	double  SpeedRef;
	double  Heading;
	double  HeadingRate;
	double  HeadingRef;
	RemusMissionStates    MissionState;
	long     Leg;
} ReconStatus_t;


class CReconDriver : public OAL::CThread
{
public:
    CReconDriver(CRemusReconImpl * pOwner,UINT32 nAddress,int nPort) ;
    virtual ~CReconDriver(void);

    bool connect();

	bool startThread();
	bool stopThread();

    bool enable();
    bool disable();

    bool enableDirectControlMode();
    bool disableDirectControlMode();

    bool enableDepthOnlyMode();
    bool disableDepthOnlyMode();

	bool isActive() { return m_bActive; }
	bool isConnected() { return m_bConnected; }
	bool isRunning() { return m_bRunning; }

	bool setCommandedHeading    ( double Heading );
	bool setCommandedAltitude   ( double Altitude );
	bool setCommandedDepth      ( double Depth );
	bool setCommandedSpeed      ( double Speed );
	bool setCommandedRpm        ( double Rpm  );
	bool setDirectControl       ( double Thrust, double Rudder, double Sternplane );
    bool setDirectControl       ( double Thrust, double Rudder, double Port, double Starboard );
	bool setDirectControlFwd	( double Rudder, double Port,   double Starboard);
	bool setCurrentWaypoint     ( double Latitude, double Longitude );
	bool setReturnToClosesPoint	( int ReturnToClosest);

	bool getRemoteControlState  ( ReconStatus_t & stat);
	bool setRemoteControlState (ReconStatus_t & stat);

private:
    void main( void );
	void handleMsg(char * pBuffer);
	bool decodeCmdAck( char * pBuffer );
	void decodeVersion( char * pBuffer );
	void decodeStateMsg(char * pBuffer);
	double decodeLongitude(char * pBuffer );
	double decodeLatitude(char * pBuffer );
	void sendVersion();
	bool verifyMsg(char * pBuffer, int nLen);
	void send(const char * szStr );

    bool						m_bActive;
	bool						m_bRunning;
    bool						m_bDirect;
	bool						m_bConnected;
	bool						m_bGotVersion;
	UINT32						m_nReconAddress;
	int							m_nReconPort;
	sockaddr_in					m_Dest;
    CRemusReconImpl				*m_pOwner;
	OAL::CAutoResetEvent		m_CmdAck;
	ReconStatus_t				m_State;
	CBerkeleySocket				m_Socket;
	long						m_nVerMaj;
	long						m_nVerMin;
	int							m_nThrust;
	int							m_nPlane[6];

	OAL::CCriticalSection m_mutex;
};
