
#include "RemusReconImpl.h"

#include "CommonCode/Log/LogManager.h"
#include "CommonCode/RTOS abstraction layer/config.h"

#ifndef PI
#define PI 3.14159265
#endif

// REMOVE WHEN REMUS SIMULATOR IS AVAILABLE
#ifdef __DEBUGTHREAD
	#include "DebugThread.h"
#endif

CRemusReconImpl::CRemusReconImpl(void) :
	m_pEventHandler(NULL),
	m_pGuidanceThread(NULL),
	m_EventLog(getName())
{
	char pIP[128];

	// Get RECON UDP port
	int nPort = GetConfigIntFromFile(getConfigFileName(), "Communication", "UdpPort", 23456);
	
	// Get RECON IP address
	GetConfigStringFromFile(getConfigFileName(), "Communication", "ReconIP", "127.0.0.1", pIP, sizeof(pIP));
	
	// Create RECON driver
	m_pRCDriver = new CReconDriver(this, inet_addr(pIP), nPort);

	// Create guidance thread
	m_pGuidanceThread = new CGuidanceThread(this);

	// Create sideScan thread
	m_pSideScanThread = new CSideScanThread(this);

	// Create nav thread
	m_pNavThread = new CNavThread(this);

	// Create magnetometer thread
	m_pMagnetometerThread = new CMagnetometerThread(this);

	// Create sensor fusion thread
	m_pSensorFusionThread = new CSensorFusionThread(this);

	// Create debug thread if debugging
	#if(defined(__DEBUGTHREAD))
		// Create a new debug thread object
		std::cout << "DebugThread created" << std::endl;
		m_debugThread = new CDebugThread(this);
	#endif
}

CRemusReconImpl::~CRemusReconImpl(void)
{
	if(m_pGuidanceThread != NULL)
	{
		delete(m_pGuidanceThread);
		m_pGuidanceThread = NULL;
	}

	if(m_pRCDriver != NULL)
	{
		delete(m_pRCDriver);
		m_pRCDriver = NULL;
	}

	// Delete debug thread
	#if(defined(__DEBUGTHREAD))
		if(m_debugThread != NULL)
		{
			// Stop and delete the debug thread object
			m_debugThread->stopThread();
			delete(m_debugThread);
		}
	#endif

	// Stop and delete the SideScan thread object
	if(m_pSideScanThread != NULL)
	{
		m_pSideScanThread->stopThread();
		delete(m_pSideScanThread);
	}

	// Stop and delete the nav thread object
	if(m_pNavThread != NULL)
	{
		m_pNavThread->stopThread();
		delete(m_pNavThread);
	}

	// Stop and delete the magnetometer thread object
	if(m_pMagnetometerThread != NULL)
	{
		m_pMagnetometerThread->stopThread();
		delete(m_pMagnetometerThread);
	}

	// Stop and delete the sensor fusion thread object
	if(m_pSensorFusionThread != NULL)
	{
		m_pSensorFusionThread->stopThread();
		delete(m_pSensorFusionThread);
	}
}

void CRemusReconImpl::activate ()
{
	try
	{	
		//Check if powered on
		if(!hasPower())
			return;

		if(isActive())
			return;

		//Call the super
		CPayloadSensor::activate();

		#if(defined(__DEBUGTHREAD))
			// Start debug thread
			m_debugThread->startThread();
		#endif

		// Start SideScan thread
		m_pSideScanThread->startThread();

		// Start nav thread
		m_pNavThread->startThread();

		// Start magnetometer thread
		m_pMagnetometerThread->startThread();

		// Start RECON driver
		m_pRCDriver->startThread();

		// Start guidance thread
		m_pGuidanceThread->startThread();
		
		// Start sensor fusion thread
		m_pSensorFusionThread->startThread();

		// Logging on
		LoggingOn();

		//Notify PPOS and OS
		sendEventToOs(eRCEventOperationOn);
	}
	catch(...)
	{
		CLogManager::error("CableTrackerPlugin::CRemusReconImpl","activate","Error when sending activate command to Command handler.");
	}
}

int CRemusReconImpl::doCommands (const void * sendData, int len)
{
	int nCmdCnt = 0;

	const unsigned char *pCmd = (reinterpret_cast<const unsigned char *>(sendData));

	while (len > 0)
	{
		bool bOk = false;
		RCCmdId nCmd = (RCCmdId)*pCmd++; --len;
		unsigned int nPar1, nPar2, nPar3, nPar4;

		if (nCmd >= eRCCmdLastCmd)
		{
			sendReplyToOsCmdStatus( eRCCRCmdUnknown, nCmd );
			return nCmdCnt;
		}

		++nCmdCnt;

		switch(nCmd)
		{
		case eRCGetRemoteControlStatus:
			{
				ReconStatus_t stat;
				if (m_pRCDriver->getRemoteControlState(stat))
				{
					CLogManager::log("CRemusReconImpl","Status: ", "%s %s %s %s %s %d %d %d %.1f %.1f %.1f %.1f %.1f %.1f %.1f",
						stat.RemoteControlAllowed ? "Allowed" : "Not-Allowed",
						stat.RemoteControlEnabled ? "Enabled" : "Disabled",
						stat.RemoteControlActive  ? "Active" : "InActive",
						stat.DirectControlEnabled ? "Direct" : "Non-Direct",
						stat.DepthOnlyControlEnabled ? "DepthOnly" : "Non-DepthOnly",
						stat.Heading,
						stat.Speed,
						stat.Altitude,
						stat.Depth,
						stat.Pitch,
						stat.Roll,
						stat.Rpm );
					bOk = true;
				}
			}
			break;

		case eRCEnable:
			bOk = m_pRCDriver->enable();
			break;
		case eRCDisable:
			bOk = m_pRCDriver->disable();
			break;

		case eRCEnableDepthOnly:
			bOk = m_pRCDriver->enableDepthOnlyMode();
			break;
		case eRCDisableDepthOnly:
			bOk = m_pRCDriver->disableDepthOnlyMode();
			break;

		case eRCEnableDirect:
			bOk = m_pRCDriver->enableDirectControlMode();
			break;
		case eRCDisableDirect:
			bOk = m_pRCDriver->disableDirectControlMode();
			break;

		case eRCSetHeading:
			if (len < 1)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			bOk = m_pRCDriver->setCommandedHeading(2*(double)nPar1*PI/180.0);     // parameter is deg/2
			break;

		case eRCSetAltitude:
			if (len < 1)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			bOk = m_pRCDriver->setCommandedAltitude((double)nPar1);     // parameter is m
			break;

		case eRCSetDepth:
			if (len < 1)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			bOk = m_pRCDriver->setCommandedDepth((double)nPar1);     // parameter is m
			break;

		case eRCSetSpeed:
			if (len < 1)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			bOk = m_pRCDriver->setCommandedSpeed((double)nPar1/10.0);     // parameter is 10*m/s
			break;

		case eRCSetRpm:
			if (len < 1)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			bOk = m_pRCDriver->setCommandedRpm((double)nPar1);     // parameter is Rpm
			break;

		case eRCSetDirect2:
			if (len < 3)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			nPar2 = *pCmd++; --len;
			nPar3 = *pCmd++; --len;
			bOk = m_pRCDriver->setDirectControl((double)nPar1,(double)nPar2,(double)nPar3);     // parameters are thrust, rudder, sternplane in %
			break;

		case eRCSetDirect3:
			if (len < 4)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			nPar2 = *pCmd++; --len;
			nPar3 = *pCmd++; --len;
			nPar4 = *pCmd++; --len;
			bOk = m_pRCDriver->setDirectControl((double)nPar1,(double)nPar2,(double)nPar3,(double)nPar4);     // parameters are thrust, rudder, port, starboard in %
			break;

		case eRCSetWaypoint:
			if (len < 4)
			{
				sendReplyToOsCmdStatus( eRCCRCmdError, nCmd );
				return nCmdCnt;
			}
			nPar1 = *pCmd++; --len;
			nPar1 += 256 * *pCmd++; --len;
			nPar2 = *pCmd++; --len;
			nPar2 += 256 * *pCmd++; --len;
			bOk = m_pRCDriver->setCurrentWaypoint((double)nPar1/10000.0,(double)nPar2/10000.0);     // parameter is 10000*rad
			break;
		}
		if (bOk)
			sendReplyToOsCmdStatus( eRCCRCmdExecuted, nCmd );
		else
			sendReplyToOsCmdStatus( eRCCRCmdFailed, nCmd );
	}

	return nCmdCnt;
}

void CRemusReconImpl::LoggingOff ()
{
	//Check if powered on
	if(!hasPower())
		return;

	CPayloadSensor::LoggingOff();

	sendEventToOs(eRCEventLoggingOff);
}

void CRemusReconImpl::LoggingOn ()
{
	//Check if powered on
	if(!hasPower())
		return;

	CPayloadSensor::LoggingOn();

	sendEventToOs(eRCEventLoggingOn);
}

int CRemusReconImpl::payloadData (void* recvData, int len)
{
	recvData;
	len;
	return 0;
}

void CRemusReconImpl::powerOff ()
{
	//Check if powered on
	if(!hasPower())
		return;

	//Check if active
	if(isActive())
		stop();

	//Call the super
	CPayloadSensor::powerOff();

	sendEventToOs(eRCEventPowerOff);
}

void CRemusReconImpl::powerOn ()
{
	//Check if powered on
	if(hasPower())
		return;

	//Call the super
	CPayloadSensor::powerOn();

	sendEventToOs(eRCEventPowerOn);

	if (!m_pRCDriver->connect())
		sendEventToOs(eRCEventConnectionFailed);
}

void CRemusReconImpl::setEventHandler (CPluginEventHandler* pEventHandler)
{
	m_pEventHandler = pEventHandler;
}

// This function is called by PP basis to set up the plugin data storer
void CRemusReconImpl::setDataStorer (CDataStorer& storer, int storerNo)
{
	try
	{
		if (storerNo == 0)
		{
			m_pDataStorer = &storer;
			assert(m_pDataStorer != NULL);
			m_dataStorageClient.registerAsClient(*m_pDataStorer);
		}
		else if (storerNo == 1)
		{
			m_pSideScanDataStorer = &storer;
			assert(m_pSideScanDataStorer != NULL);
			m_sideScanStorageClient.registerAsClient(*m_pSideScanDataStorer);
		}
		else if (storerNo == 2)
		{
			m_pSensorFusionDataStorer = &storer;
			assert(m_pSensorFusionDataStorer != NULL);
			m_sensorFusionStorageClient.registerAsClient(*m_pSensorFusionDataStorer);
		}
		else if (storerNo == 3)
		{
			m_pMagnetometerDataStorer = &storer;
			assert(m_pMagnetometerDataStorer != NULL);
			m_magnetometerStorageClient.registerAsClient(*m_pMagnetometerDataStorer);
		}
	}
	catch (...)
	{
		std::cout << "CableTrackerPlugin::CRemusReconImpl setDataStorer: Caught exception while executing. Storer no: " << storerNo << std::endl;
		CLogManager::error("CableTrackerPlugin::CRemusReconImpl", "setDataStorer", "Caught exception while executing.");
	}
}

// This function is called by PP basis to set up the plugin event handler.
// Data stored in the event data storer ends up in the PP event log.
void CRemusReconImpl::setEventStorer (CDataStorer& storer)
{
	try
	{
		// Save and register the event data storer with the data storage client
		m_pEventDataStorer = &storer;
		assert(m_pEventDataStorer != NULL);
		m_eventStorageClient.registerAsClient(*m_pEventDataStorer);
	}
	catch (...)
	{
		CLogManager::error("CableTrackerPlugin::PluginImpl", "setEventStorer", "Caught exception while executing.");
	}
}

void CRemusReconImpl::setMissionInfo (const char* path)
{
	path;
}

void CRemusReconImpl::stop ()
{
	//Check if powered on
	if(!hasPower())
		return;

	//Check if active
	if(!isActive())
		return;

	try
	{
		CPayloadSensor::stop();

		// Stop guidance thread
		m_pGuidanceThread->stopThread();

		//Send command to plugin command handler
		if (!m_pRCDriver->disable())
			CLogManager::error("CableTrackerPlugin::CRemusReconImpl","stop","Failed to disable Remote Control interface");

		// Stop RECON thread
		m_pRCDriver->stopThread();

		//Notify PPOS and OS
		sendEventToOs(eRCEventOperationOff);

		//Turn off logging
		LoggingOff();
	}
	catch(...)
	{
		CLogManager::error("CableTrackerPlugin::CRemusReconImpl","stop","Error when sending stop command to Command handler.");
	}
}

void CRemusReconImpl::stopped ()
{
	//Check if powered on
	if(!hasPower())
		return;

	sendEventToOs(eRCEventNetworkError);
}


void CRemusReconImpl::newLine (const char* tag)
{
	tag;
}

void CRemusReconImpl::sendReplyToOsCmdStatus (RCCmdResponseId eResponseId, RCCmdId eCmdID)
{
	unsigned char msgData[2];

	msgData[0] = (unsigned char)eResponseId;
	msgData[1] = (unsigned char)eCmdID;

	try
	{
		char szLogFileText[150];				
		sprintf( szLogFileText, "Reply to OS - %d - %d", eResponseId, eCmdID );
		m_EventLog.logEvent( szLogFileText );

		m_pEventHandler->cmdResponse( msgData, 2);
	}
	catch (...) 
	{
		CLogManager::error( "CableTrackerPlugin::CRemusReconImpl", "sendReplyToOsCmdStatus", "Unable to send reply to OS on command status" );
	}
}

// Store an event string in the PP vent log
void CRemusReconImpl::storeEventString(const char *pText)
{
	try
	{
		assert( strlen(pText) < 480);
		char szLogFileText[500];

		CAbsTime TimeStamp = CHPTimer::getInstance()->getAbsTime();
		TimeStamp.getAsDateTime (szLogFileText);

		sprintf(szLogFileText,"%s %s: ", szLogFileText, getName());
		strcat(szLogFileText, pText);	
		strcat(szLogFileText,"\n");
		m_eventStorageClient.storeString( szLogFileText );
	}
	catch (...)
	{
		// Log exception to pptest.txt
		CLogManager::error("CableTrackerPlugin::CRemusReconImpl", "storeEventString", "Caught exception while executing.");
	}
}

void CRemusReconImpl::sendEventToOs (RCEventId eEventId)
{
	try
	{
		char szLogFileText[250];				
		sprintf(szLogFileText, "Event - %s - type = %d", eventToString(eEventId), (unsigned long)eEventId);
		m_EventLog.logEvent( szLogFileText );
		m_pEventHandler->event((unsigned char)eEventId);
	}
	catch(...)
	{
		CLogManager::error( "CableTrackerPlugin::CRemusReconImpl", "sendEventToOS", "Error when reporting an event to POS." );
	}
}

char * CRemusReconImpl::eventToString (RCEventId id)
{
	static char str[40];
	switch (id) {
		 case eRCEventConnectionFailed:		return "UDP socket setup failed";
		 case eRCEventOperationOff:		    return "Stopped";
		 case eRCEventOperationOn:	        return "Activated";
		 case eRCEventPowerOn:	            return "Power on";
		 case eRCEventPowerOff:	            return "Power off";
		 case eRCEventLoggingOn:		    return "Logging on";
		 case eRCEventLoggingOff:		    return "Logging off";
		 case eRCEventEnableFailed:		    return "Remote Control enable failed";
		 case eRCEventNetworkError:		    return "Network error";

		 default:
			 sprintf(str, "Event %i", id);
			 return str;
	}
}

const char* CRemusReconImpl::getCurrentFileName()
{
	std::string fName = m_pDataStorer->getCurrentFileName();
	return fName.c_str();
}

// Store data using the plugin data storer
bool CRemusReconImpl::storeData(unsigned char *pData, size_t nDataLen)
{
	// Do not store data if logging is OFF
	if(!isLogging())
	{
		return false;
	}

	if (m_pDataStorer != NULL)
	{
		// Make storeObject ThreadSafe
		OAL::CCriticalSectionLock Lock( m_dataStorageClient.getStorerCS() );

		// Reserve buffer for the data
		unsigned char* pbyStorageBuffer = m_dataStorageClient.reserveBuffer((long) nDataLen );
		if (pbyStorageBuffer != NULL)
		{
			// Copy the data to the buffer
			memcpy(pbyStorageBuffer, pData, nDataLen);

			// Instruct the storage client to store the data to file
			m_dataStorageClient.storeBuffer();

			return true;
		}
	}
	return false;
}

