#include "RemusReconImpl.h"
#include "SensorFusionThread.h"


#include "CommonCode/Log/LogManager.h"

CSensorFusionThread::CSensorFusionThread(CRemusReconImpl *pParent) :
OAL::CThread("CableTrackerPluginSensorFusionThread"),
//m_mutex("SensorFusionMutex"),
m_bRunning(false),
m_pParent(pParent),
m_totSensors(3),
m_sensorsUsed(0),
m_numStates(2),
m_bBadSetup(false),
m_bFuseSS(false),
m_bFuseMag(false),
m_bFuseAprDat(false),
m_bMultiRateEKF(true),
m_timeStep(10),
m_bHeaderCreated(false),
m_prevSideScanUpdate(0),
m_prevMagnetometerUpdate(1),
m_prevWpUpdate(1),
m_bWaitForInput(true)
//firstTimeStep(0)
{
	;
}

CSensorFusionThread::~CSensorFusionThread()
{
	;
}

void CSensorFusionThread::main()
{
	while( !m_ShutDown.wait(0) )
	{
		sleep(m_timeStep);
		if (m_pParent->getReconDriver()->isActive())
		{
			if (m_bMultiRateEKF)
				multiRateEKF();
			else
				ekfStep();
		}
	}
}

bool CSensorFusionThread::startThread()
{
	
	// Check if module has been initialized, if not, initialize module now
	if(!initializeEkf())
	{
		// Log the event to the event log
		m_pParent->storeEventString("CSensorFusionThread: Failed to initialize module.");
		std::cout << "ERROR: CableTrackerPlugin: CSensorFusionThread::startThread() Unable to initialize" << std::endl;
		return false;
	}

	try
	{
		// Start thread
		start();
		m_bRunning = true;

		// Log an event to the event log
		m_pParent->storeEventString("SensorFusionThread: Thread started.");
		return m_bRunning;
	}
	catch ( ... )
	{
		m_pParent->storeEventString("SensorFusionThread: Exception when trying to start thread.");
		return false;
	}
}

void CSensorFusionThread::stopThread()
{
	// Shutdown thread
	(void) shutDown();

	// Log an event to the event log
	m_pParent->storeEventString("SensorFusionThread: Thread stopped.");

	m_bRunning = false;
}

bool CSensorFusionThread::initializeEkf()
{
	GetConfigStringFromFile(m_pParent->getConfigFileName(), "SensorFusion", "MatrixFolder", "./", m_matrixFolder, sizeof(m_matrixFolder));

	m_bMultiRateEKF = GetConfigIntFromFile(m_pParent->getConfigFileName(), "SensorFusion", "UseMultiRateEKF", 1) == 1;
	m_timeStep = GetConfigIntFromFile(m_pParent->getConfigFileName(), "SensorFusion", "FilterTimeStep", 10);
	m_numStates = GetConfigIntFromFile(m_pParent->getConfigFileName(), "SensorFusion", "NumberOfStates", 2);

	m_bFuseSS = GetConfigIntFromFile(m_pParent->getConfigFileName(), "SensorFusion", "FuseSideScan", 1) == 1;
	m_bFuseMag = GetConfigIntFromFile(m_pParent->getConfigFileName(), "SensorFusion", "FuseMagnetometer", 1) == 1;
	m_bFuseAprDat = GetConfigIntFromFile(m_pParent->getConfigFileName(), "SensorFusion", "FuseAprioriData", 1) == 1;

	m_sensorsUsed = m_bFuseSS + m_bFuseMag + m_bFuseAprDat;

	if (m_sensorsUsed < 1)
	{
		std::cout << "CableTracker:SensorFusion: No sensor input included in config" << std::endl;
		return false;
	}

	fuseSensor.resize(m_totSensors, false);
	sensorNum.resize(m_totSensors, -1);

	fuseSensor.at(0) = m_bFuseSS;
	fuseSensor.at(1) = m_bFuseMag;
	fuseSensor.at(2) = m_bFuseAprDat;

	int count = 0;

	for (int i=0; i< m_totSensors; i++)
	{
		if (fuseSensor.at(i))
		{
			sensorNum.at(i) = count; 
			count++;
		}
	}
	
	//m_xHat = Eigen::MatrixXd::Zero(m_numStates,1);
	x_hat_prior.resize(m_numStates,1);
	y.resize(m_numStates*m_sensorsUsed,1);
	H.resize(m_numStates*m_sensorsUsed,m_numStates);
	R.resize(m_numStates*m_sensorsUsed,m_numStates*m_sensorsUsed);
	Q.resize(m_numStates,m_numStates);
	E.resize(m_numStates,m_numStates);

	P_prior = Eigen::MatrixXd::Identity(m_numStates, m_numStates);
	
	readMatrixFromFile(R,"R.txt");
	readMatrixFromFile(H,"H.txt");
	readMatrixFromFile(Q,"Q.txt");
	readMatrixFromFile(E,"E.txt");
	readMatrixFromFile(x_hat_prior,"x0.txt");
	m_xHat = x_hat_prior;

	prevTimeStep = CHPTimer::getInstance()->getAbsTime();
	prevTimeStep_ = CHPTimer::getInstance()->getAbsTime();

	return true;
}

//read space separated matrix from file
bool CSensorFusionThread::readMatrixFromFile(Eigen::MatrixXd& matrix, const char *filename)
{
	//bool readSuccess = false;
	char file [128];
	char *pEnd;
	strcpy(file, m_matrixFolder);
	
	// ensure that path to folder has a slash at the end
	pEnd = (char*) memchr(m_matrixFolder,'\0',128);
	if (*(pEnd-1) != '/' && *(pEnd-1) != '\\')
		strcat(file, "\\" );

	strcat(file, filename);

	std::ifstream fin (file);

	if (fin.is_open())
	{
		int row = 0;
		int col = 0;
		float item;

		while (fin >> item)
		{
			if (row >= matrix.rows() || col >= matrix.cols())
				{
					std::cout << "SensorFusion: Error, expected a " << matrix.cols() << "x" << matrix.rows() << " matrix in " << filename  << std::endl;
					m_bBadSetup = true;
					break;
				}
			else
				matrix(row, col) = item;

			col++;
			if (col >= matrix.cols())
			{
				row++;
				col = 0;
			}
		}

		if (row != matrix.rows() || col != 0)
		{
			std::cout << "SensorFusion: Error, expected a " << matrix.cols() << "x" << matrix.rows() << " matrix in " << filename  << std::endl;
			m_bBadSetup = true;
		}
	}
	else if (filename == "x0.txt")
		for (int i = 0; i < m_numStates; i++)
			x_hat_prior(i,0) = 0;
	else
	{
		std::cout << "SensorFusion: Unable to open " << filename << std::endl;
		m_bBadSetup = true;
	}

	//std::cout << matrix << std::endl;

	return !m_bBadSetup;
}

//perform one extended kalman filter step
void CSensorFusionThread::ekfStep()
{
	ReconStatus_t reconState;
	m_pParent->getReconDriver()->getRemoteControlState(reconState);

	if (!m_bHeaderCreated)
	{
		createDataHeader();
	}

	Eigen::MatrixXd Phi = Eigen::MatrixXd::Identity(m_numStates,m_numStates);
	Eigen::MatrixXd K;
	Eigen::MatrixXd r;
	
	Eigen::MatrixXd P;
	Eigen::MatrixXd Gamma;

	// Update y vector
	getMeasurements();

	CAbsTime timeStamp = CHPTimer::getInstance()->getAbsTime();
	double h = (double) (timeStamp - prevTimeStep)/1000000;
	prevTimeStep = CHPTimer::getInstance()->getAbsTime();
	double speed = reconState.Speed;
	Gamma = h * E;

	//Kalman gain
	K = P_prior*H.transpose() * (H*P_prior*H.transpose() + R).inverse();

	//State estimate update
	r = y-H*x_hat_prior;
	m_xHat = x_hat_prior + K*r;

	double auvHeading = reconState.Heading;
	double relativeHeading = m_xHat(1,0) - auvHeading;
	Phi(0,1) += h*speed * cos(relativeHeading);

	//Error covariance
	P = (Eigen::MatrixXd::Identity(m_numStates,m_numStates) - K*H) * P_prior * (Eigen::MatrixXd::Identity(m_numStates,m_numStates) - K*H).transpose() + K*R*K.transpose();
	
	//Projection
	Eigen::MatrixXd f(2,1);
	f << speed*sin(relativeHeading), 0;

	P_prior = Phi*P*Phi.transpose() + Gamma*Q*Gamma.transpose();
	x_hat_prior = m_xHat + h*f;

	std::cout << "x_hat_prior: \n" << x_hat_prior << std::endl;

	//if (firstTimeStep == 0)
	//	firstTimeStep = timeStamp;

	double time = (double) timeStamp;
	double leg = (double) reconState.Leg;
	double wp = (double) m_pParent->getGuidanceThread()->getGuidanceImpl()->getCurrentWaypoint();
	double LOS_heading = m_pParent->getGuidanceThread()->getGuidanceImpl()->getHeadingCommandRadians();
	
	double earthRadius = 6371000.0;
	double cableLat = reconState.Latitude - m_xHat(0,0)*sin(m_xHat(1,0))/earthRadius;
	double cableLong = reconState.Longitude + m_xHat(0,0)*cos(m_xHat(1,0))/earthRadius;

	storeSensorFusionData((unsigned char *) &time, sizeof(time));
	storeSensorFusionData((unsigned char *) m_xHat.data(), m_numStates*sizeof(m_xHat(0,0)));
	storeSensorFusionData((unsigned char *) r.data(), 2*m_sensorsUsed*sizeof(r(0,0)));
	storeSensorFusionData((unsigned char *) y.data(), 2*m_sensorsUsed*sizeof(y(0,0)));
	storeSensorFusionData((unsigned char *) &auvHeading, sizeof(auvHeading));
	storeSensorFusionData((unsigned char *) &speed, sizeof(speed));
	storeSensorFusionData((unsigned char *) &reconState.Latitude, sizeof(reconState.Latitude));
	storeSensorFusionData((unsigned char *) &reconState.Longitude, sizeof(reconState.Longitude));
	storeSensorFusionData((unsigned char *) &reconState.Depth, sizeof(reconState.Depth));
	storeSensorFusionData((unsigned char *) &reconState.Altitude, sizeof(reconState.Altitude));
	storeSensorFusionData((unsigned char *) &reconState.Pitch, sizeof(reconState.Pitch));
	storeSensorFusionData((unsigned char *) &reconState.Roll, sizeof(reconState.Roll));
	storeSensorFusionData((unsigned char *) &reconState.Heading, sizeof(reconState.Heading));
	storeSensorFusionData((unsigned char *) &leg, sizeof(leg));
	storeSensorFusionData((unsigned char *) &wp, sizeof(wp));
	storeSensorFusionData((unsigned char *) &LOS_heading, sizeof(LOS_heading));
	storeSensorFusionData((unsigned char *) &cableLat, sizeof(cableLat));
	storeSensorFusionData((unsigned char *) &cableLong, sizeof(cableLong));

	//std::cout << Phi << std::endl;
	//std::cout << P << std::endl;
	//std::cout << Gamma << std::endl;
	//std::cout << Q << std::endl;

	std::cout << "x_hat: " << m_xHat(0,0) << "   " << m_xHat(1,0)*180/3.141592 << std::endl;// << "  time =  " << timeStamp - firstTimeStep << " micro sec "  << std::endl;
}

bool CSensorFusionThread::createDataHeader()
{
	int success = 0;
	double numSensors = (double) m_sensorsUsed;
	double numStates = (double) m_numStates;

	success += storeSensorFusionData((unsigned char *) &numSensors, sizeof(numSensors));
	success += storeSensorFusionData((unsigned char *) &numStates, sizeof(numStates));

	double fuseSS = (double) m_bFuseSS;
	double fuseMag = (double) m_bFuseMag;
	double fuseAprDat = (double) m_bFuseAprDat;
	
	success += storeSensorFusionData((unsigned char *) &fuseSS, sizeof(fuseSS));
	success += storeSensorFusionData((unsigned char *) &fuseMag, sizeof(fuseMag));
	success += storeSensorFusionData((unsigned char *) &fuseAprDat, sizeof(fuseAprDat));

	if (success == 5)
		m_bHeaderCreated = true;
	else
		std::cout << "CSensorFusionThread: failed to create data header" << std::endl;

	return m_bHeaderCreated;
}

// get measurements and insert into y vector
void CSensorFusionThread::getMeasurements()
{
	//Get measurements
	double crossTrackErrSS =  generateGaussianNoise(1, 0.001); //m_pParent->getSideScanThread()->getCableXPos();
	double headingSS = generateGaussianNoise(1, 0.001); //m_pParent->getSideScanThread()->getcableHeadingRad();

	double crossTrackErrMag= generateGaussianNoise(1.1, 0.001);
	double headingMag = generateGaussianNoise(1.1, 0.001);

	//Apriori data
	double crossTrackErrWP = m_pParent->getGuidanceThread()->getGuidanceImpl()->getWpCrossTrackError(); // generateGaussianNoise(0.9, 0.001);
	double headingWP = m_pParent->getGuidanceThread()->getGuidanceImpl()->getWpCableHeading();  //headingWP;

	std::vector<double> crossTrackError(m_totSensors);
	std::vector<double> cableHeading(m_totSensors);

	crossTrackError.at(0) = crossTrackErrSS;
	crossTrackError.at(1) = crossTrackErrMag;
	crossTrackError.at(2) = crossTrackErrWP;

	cableHeading.at(0) = headingSS;
	cableHeading.at(1) = headingMag;
	cableHeading.at(2) = headingWP;

	int j = 0;

	for (int i = 0; i<m_totSensors; i++)
	{
		if (fuseSensor.at(i))
		{
			y(2*j,0) = crossTrackError.at(i);
			y(1+2*j,0) = cableHeading.at(i);
			j++;
		}
	}
}

void CSensorFusionThread::multiRateEKF()
{
	if (!m_bHeaderCreated)
		createDataHeader();

	double crossTrackErrWP = m_pParent->getGuidanceThread()->getGuidanceImpl()->getWpCrossTrackError(); // generateGaussianNoise(0.9, 0.001);
	double headingWP = m_pParent->getGuidanceThread()->getGuidanceImpl()->getWpCableHeading();  //headingWP;
	ReconStatus_t reconState;
	m_pParent->getReconDriver()->getRemoteControlState(reconState);

	std::vector<bool> sampled(m_sensorsUsed, false);	//input true for new samples
	std::vector<double> crossTrackError(m_totSensors);
	std::vector<double> cableHeading(m_totSensors);

	//sidescan
	if (fuseSensor.at(0) && m_pParent->getSideScanThread()->getMessageNumber() > m_prevSideScanUpdate)
	{
		m_prevSideScanUpdate = m_pParent->getSideScanThread()->getMessageNumber();
		sampled.at(sensorNum.at(0)) = true;
		crossTrackError.at(sensorNum.at(0)) =  m_pParent->getSideScanThread()->getCableCrossTrack(); //   generateGaussianNoise(1, 0.001);//
		cableHeading.at(sensorNum.at(0)) = m_pParent->getSideScanThread()->getCableHeadingRad(); //  generateGaussianNoise(1, 0.001); //
	}

	bool preciseMeasurement = m_pParent->getMagnetometerThread()->isMeasurementPrecise();

	//magnetometer
	if (fuseSensor.at(1) && m_pParent->getMagnetometerThread()->getMessageNumber() > m_prevMagnetometerUpdate)
	{
		m_prevMagnetometerUpdate =  m_pParent->getMagnetometerThread()->getMessageNumber();
		sampled.at(sensorNum.at(1)) = true;
		crossTrackError.at(sensorNum.at(1)) = m_pParent->getMagnetometerThread()->getCableCrossTrack() + generateGaussianNoise(0, 0.5) * preciseMeasurement;
		cableHeading.at(sensorNum.at(1)) =	m_pParent->getMagnetometerThread()->getCableHeadingRad() + generateGaussianNoise(0, 0.005) * preciseMeasurement;
	}

	//sample apriori data if included
	if (fuseSensor.at(2) && m_pParent->getGuidanceThread()->getGuidanceImpl()->getWpLOSNumber() > m_prevWpUpdate)
	{
		//if (std::accumulate(sampled.begin(), sampled.end(), 0) < 1)
		{
			int currentWP = m_pParent->getGuidanceThread()->getGuidanceImpl()->getCurrentWaypoint();

			//double random =  (double) rand()/RAND_MAX; //for simulation

			//if (currentWP !=8 && random < 0.7)
			{
				sampled.at(sensorNum.at(2)) = true;
				crossTrackError.at(sensorNum.at(2)) = crossTrackErrWP;
				cableHeading.at(sensorNum.at(2)) = headingWP;
				m_prevWpUpdate++;
			}
		}
	}

	int sampledSensors = std::accumulate(sampled.begin(), sampled.end(), 0);

	if (sampledSensors == 0 && m_bWaitForInput)  //sampledSensors - sampled.at(sensorNum.at(1))
	{
		return;
	}

	m_bWaitForInput = false;

	Eigen::MatrixXd y_(2*sampledSensors,1);
	Eigen::MatrixXd H_(2*sampledSensors,m_numStates);
	Eigen::MatrixXd R_ = Eigen::MatrixXd::Zero(2*sampledSensors,2*sampledSensors);

	int j = 0;
	int s = m_numStates;
	y = Eigen::MatrixXd::Zero(m_numStates*m_sensorsUsed,1);

	//Multi-rate output sampler
	for (int i=0; i < m_totSensors; i++)
	{
		if (sensorNum.at(i) == -1)
			continue;
		//only append values in y, H and R for sampled sensors
		if (sampled.at(sensorNum.at(i)))
		{
			y_(2*j,0) = crossTrackError.at(sensorNum.at(i));
			y_(1+2*j,0) = cableHeading.at(sensorNum.at(i));

			H_.block(j*s,0,s,s) = H.block(j*s,0,s,s);
			R_.block(j*s,j*s,s,s) = R.block(j*s,j*s,s,s);	//assuming that only the diagonal in R is used

			j++;
		}
	}

	//std::cout << "y_: \n" << y_ << std::endl;
	//std::cout << "H_: \n" << H_ << std::endl;
	//std::cout << "R_: \n" << R_ << std::endl;

	Eigen::MatrixXd Phi = Eigen::MatrixXd::Identity(m_numStates,m_numStates);
	Eigen::MatrixXd K;
	Eigen::MatrixXd r;
	
	Eigen::MatrixXd P;
	Eigen::MatrixXd Gamma;
	Eigen::MatrixXd f(2,1);

	CAbsTime timeStamp = CHPTimer::getInstance()->getAbsTime();
	double h = (double) (timeStamp - prevTimeStep_)/1000000;
	prevTimeStep_ = CHPTimer::getInstance()->getAbsTime();
	double speed = reconState.Speed;
	Gamma = h * E;

	//Update
	if (sampledSensors > 0)
	{
		K = P_prior*H_.transpose() * (H_*P_prior*H_.transpose() + R_).inverse();
		r = y_- H_*x_hat_prior;
		m_xHat = x_hat_prior + K*r;
		P = (Eigen::MatrixXd::Identity(s,s) - K*H_) * P_prior * (Eigen::MatrixXd::Identity(s,s) - K*H_).transpose() + K*R_*K.transpose();
	}
	else
	{
		m_xHat = x_hat_prior;
		P = P_prior;
	}
	
	double auvHeading = reconState.Heading;
	double relativeHeading = m_xHat(1,0) - auvHeading;

	Phi(0,1) += h*speed * cos(relativeHeading);
	f << speed*sin(relativeHeading), 0;

	//Projection
	P_prior = Phi*P*Phi.transpose() + Gamma*Q*Gamma.transpose();
	x_hat_prior = m_xHat + h*f;

	//if (timeStamp % 10 == 0)
	//	std::cout << "MR-EKF: " <<sampledSensors << " samples, x_hat = " << m_xHat(0,0) << "   " << m_xHat(1,0)*180/3.141592 << std::endl;

	//std::cout << "P_prior: \n" << P_prior << std::endl;

	//Store data
	if (!m_bHeaderCreated)
	{
		std::cout << "SensorFusionThread: Warning: data header not created!" << std::endl;
		return;
	}

	Eigen::MatrixXd r_tot = Eigen::MatrixXd::Zero(s*m_sensorsUsed,1);
	j = 0;
	for (int i=0; i < m_totSensors; i++)
	{
		if (sensorNum.at(i) == -1)
			continue;

		if (sampled.at(sensorNum.at(i)))
		{
			r_tot.block(sensorNum.at(i)*s,0,s,1) = r.block(j,0,s,1);
			
			y(s*sensorNum.at(i),0) = crossTrackError.at(sensorNum.at(i));
			y(1+s*sensorNum.at(i),0) = cableHeading.at(sensorNum.at(i));

			j++;
		}
	}

	double time = (double) timeStamp;
	double leg = (double) reconState.Leg;
	double wp = (double) m_pParent->getGuidanceThread()->getGuidanceImpl()->getCurrentWaypoint();
	double LOS_heading = m_pParent->getGuidanceThread()->getGuidanceImpl()->getHeadingCommandRadians();
	//double AUV_heading = reconState.Heading*PI/180;

	double earthRadius = 6371000.0;

	double cableLat = reconState.Latitude - m_xHat(0,0)*sin(m_xHat(1,0))/earthRadius * 180 / PI;
	double cableLong = reconState.Longitude + m_xHat(0,0)*cos(m_xHat(1,0))/earthRadius * 180 / PI;

	storeSensorFusionData((unsigned char *) &time, sizeof(time));
	storeSensorFusionData((unsigned char *) m_xHat.data(), m_numStates*sizeof(m_xHat(0,0)));
	storeSensorFusionData((unsigned char *) r_tot.data(), 2*m_sensorsUsed*sizeof(r_tot(0,0)));
	storeSensorFusionData((unsigned char *) y.data(), 2*m_sensorsUsed*sizeof(y(0,0)));
	storeSensorFusionData((unsigned char *) &auvHeading, sizeof(auvHeading));
	storeSensorFusionData((unsigned char *) &speed, sizeof(speed));
	storeSensorFusionData((unsigned char *) &reconState.Latitude, sizeof(reconState.Latitude));
	storeSensorFusionData((unsigned char *) &reconState.Longitude, sizeof(reconState.Longitude));
	storeSensorFusionData((unsigned char *) &reconState.Depth, sizeof(reconState.Depth));
	storeSensorFusionData((unsigned char *) &reconState.Altitude, sizeof(reconState.Altitude));
	storeSensorFusionData((unsigned char *) &reconState.Pitch, sizeof(reconState.Pitch));
	storeSensorFusionData((unsigned char *) &reconState.Roll, sizeof(reconState.Roll));
	storeSensorFusionData((unsigned char *) &reconState.Heading, sizeof(reconState.Heading));
	storeSensorFusionData((unsigned char *) &leg, sizeof(leg));
	storeSensorFusionData((unsigned char *) &wp, sizeof(wp));
	storeSensorFusionData((unsigned char *) &LOS_heading, sizeof(LOS_heading));
	storeSensorFusionData((unsigned char *) &cableLat, sizeof(cableLat));
	storeSensorFusionData((unsigned char *) &cableLong, sizeof(cableLong));

	//std::cout << "EKF: " << reconState.Latitude << "  " << reconState.Longitude << std::endl;
}


// Generate gaussian noise with Box-Muller function from https://en.wikipedia.org/wiki/Box%E2%80%93Muller_transform
double CSensorFusionThread::generateGaussianNoise(double mu, double sigma)
{
	const double epsilon = 	DBL_MIN;  //std::numeric_limits<double>::min();
	const double two_pi = 2.0*3.14159265358979323846;

	static double z0, z1;
	static bool generate;
	generate = !generate;

	if (!generate)
	   return z1 * sigma + mu;

	double u1, u2;
	do
	 {
	   u1 = rand() * (1.0 / RAND_MAX);
	   u2 = rand() * (1.0 / RAND_MAX);
	 }
	while ( u1 <= epsilon );

	z0 = sqrt(-2.0 * log(u1)) * cos(two_pi * u2);
	z1 = sqrt(-2.0 * log(u1)) * sin(two_pi * u2);
	return z0 * sigma + mu;
}

// Store data using the plugin data storer
bool CSensorFusionThread::storeSensorFusionData(unsigned char *pData, size_t nDataLen)
{
	//OAL::CCriticalSectionLock Lock(m_mutex);

	if (m_pParent->m_pSensorFusionDataStorer != NULL)
	{
		// Make storeObject ThreadSafe
		OAL::CCriticalSectionLock Lock( m_pParent->m_sensorFusionStorageClient.getStorerCS() );

		// Reserve buffer for the data
		unsigned char* pbyStorageBuffer = m_pParent->m_sensorFusionStorageClient.reserveBuffer((long) nDataLen );
		if (pbyStorageBuffer != NULL)
		{
			// Copy the data to the buffer
			memcpy(pbyStorageBuffer, pData, nDataLen);

			// Instruct the storage client to store the data to file
			m_pParent->m_sensorFusionStorageClient.storeBuffer();
			return true;
		}
		else
		{
			CLogManager::error("CSensorFusionThread", "storeSensorFusionData", "Storage buffer == NULL.");
			return false;
		}
	}
	else
	{
		CLogManager::error("CSensorFusionThread", "storeSensorFusionData", "Data storer == NULL.");
		return false;
	}
}